# Copyright 2023 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import random
import string
from dataclasses import asdict, dataclass, field
from typing import Dict, List

from linien_common.communication import PathLike, RestorableParameterValues
from linien_common.config import SERVER_PORT, USER_DATA_PATH

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def generate_random_key():
    """Generate a random key for the device."""
    return "".join(random.choice(string.ascii_lowercase) for _ in range(10))


@dataclass
class Device:
    """A device that can be connected to."""

    key: str = field(default_factory=generate_random_key)
    name: str = field(default_factory=str)
    host: str = field(default_factory=str)
    port: int = SERVER_PORT
    username: str = field(default_factory=str)
    password: str = field(default_factory=str)
    parameters: Dict[str, RestorableParameterValues] = field(default_factory=dict)

    def __post_init__(self):
        if self.host == "":
            self.host = "rp-xxxxxx.local"
        if self.username == "":
            self.username = "root"
        if self.password == "":
            self.password = "root"
        # FIXME: is this even necessary?
        if self.host in ("localhost", "127.0.0.1"):
            # RP is configured such that "localhost" doesn't point to 127.0.0.1 in all
            # cases
            self.host = "127.0.0.1"

    def __eq__(self, other):
        if isinstance(other, Device):
            return self.key == other.key
        else:
            return False


def add_device(
    device: Device, path: PathLike = USER_DATA_PATH / "devices.json"
) -> None:
    """Add a new device to the device list and save it to disk."""
    devices = load_device_list(path)
    if device in devices:
        raise KeyError(f"Device with key {device.key} already exists in {path}.")
    devices.append(device)
    save_device_list(devices, path)
    logger.debug(f"Added device with key {device.key} to {path}.")


def load_device(key: str, path: PathLike) -> Device:
    """Load a device from disk."""
    devices = load_device_list(path)
    for device in devices:
        if device.key == key:
            return device
    raise KeyError(f"Device with key {key} doesn't exist in {path}.")


def delete_device(
    device: Device, path: PathLike = USER_DATA_PATH / "devices.json"
) -> None:
    """Remove a device from the device list and save it to disk."""
    devices = load_device_list(path)
    devices.remove(device)
    save_device_list(devices, path)


def update_device(
    device: Device, path: PathLike = USER_DATA_PATH / "devices.json"
) -> None:
    """Update a device in the device list and save it to disk."""
    devices = load_device_list(path)
    if device not in devices:
        raise KeyError(f"Device with key {device.key} doesn't exist in {path}.")
    # this updates since equality is defined by the device key, see Device.__eq__
    devices[devices.index(device)] = device
    save_device_list(devices, path)
    logger.debug(f"Updated device with key {device.key} in {path}.")


def move_device(device: Device, direction: int) -> None:
    """Move a device in the device list and save it to disk."""
    devices = load_device_list()
    current_index = devices.index(device)
    new_index = current_index + direction
    devices.insert(new_index, devices.pop(current_index))
    save_device_list(devices)


def save_device_list(
    devices: List[Device], path: PathLike = USER_DATA_PATH / "devices.json"
) -> None:
    """Save a device list to disk."""
    with open(path, "w") as f:
        logger.debug(f"Saving devices to {path}.")
        json.dump({i: asdict(device) for i, device in enumerate(devices)}, f, indent=2)


def load_device_list(
    path: PathLike = USER_DATA_PATH / "devices.json",
) -> List[Device]:
    """Load the device list from disk."""
    try:
        with open(path, "r") as f:
            logger.debug(f"Loading devices from {path}.")
            devices = [Device(**value) for _, value in json.load(f).items()]
    except FileNotFoundError:
        logger.debug("No devices.json found. Return empty list.")
        devices = []
    return devices



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2023 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import sys

import linien_client
from fabric import Connection
from linien_client.exceptions import (
    InvalidServerVersionException,
    ServerNotInstalledException,
)
from linien_common.communication import hash_username_and_password

from .device import Device

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def read_remote_version(
    device: Device, ssh_port: int = 22, out_stream=sys.stdout
) -> str:
    """Read the remote version of linien."""

    if not out_stream:
        # sys.stdout is not available in the pyinstaller build, redirect it to avoid
        # AttributeError: 'NoneType' object has no attribute 'write'
        out_stream = open(os.devnull, "w")

    with Connection(
        device.host,
        user=device.username,
        port=ssh_port,
        connect_kwargs={"password": device.password},
    ) as conn:
        result = conn.run(
            'python3 -c "import linien_server; print(linien_server.__version__);"',
            out_stream=out_stream,
            err_stream=out_stream,
            warn=True,
        )
    if result.ok:
        return result.stdout.strip()
    else:
        raise ServerNotInstalledException()


def start_remote_server(
    device: Device, ssh_port: int = 22, out_stream=sys.stdout
) -> None:
    """Start the remote linien server."""

    if not out_stream:
        # sys.stdout is not available in the pyinstaller build, redirect it to avoid
        # AttributeError: 'NoneType' object has no attribute 'write'
        out_stream = open(os.devnull, "w")

    with Connection(
        device.host,
        user=device.username,
        port=ssh_port,
        connect_kwargs={"password": device.password},
    ) as conn:
        local_version = linien_client.__version__.split("+")[0]
        remote_version = read_remote_version(device, ssh_port).split("+")[0]

        if (local_version != remote_version) and not ("dev" in local_version):
            raise InvalidServerVersionException(local_version, remote_version)

        logger.debug("Sending credentials")
        conn.run(
            'python3 -c "from linien_common.communication import write_hash_to_file;'
            f"write_hash_to_file('{hash_username_and_password(device.username, device.password)}')\"",  # noqa E501
            out_stream=out_stream,
            err_stream=out_stream,
            warn=True,
        )

        logger.debug("Starting server")
        conn.run(
            "linien-server start",
            out_stream=out_stream,
            err_stream=out_stream,
            warn=True,
        )


def install_remote_server(
    device: Device, ssh_port: int = 22, out_stream=sys.stdout
) -> None:
    """Install the remote linien server."""

    if not out_stream:
        # sys.stdout is not available in the pyinstaller build
        out_stream = open(os.devnull, "w")

    with Connection(
        device.host,
        user=device.username,
        port=ssh_port,
        connect_kwargs={"password": device.password},
    ) as conn:
        local_version = linien_client.__version__.split("+")[0]
        cmds = [
            "linien-server stop",
            "pip3 uninstall linien-server -y",
            "pip3 uninstall linien-common -y",
            f"pip3 install linien-server=={local_version} --no-cache-dir",
        ]
        for cmd in cmds:
            out_stream.write(f">> {cmd}\n")
            result = conn.run(
                cmd, out_stream=out_stream, err_stream=out_stream, warn=True
            )
            if result.ok:
                logger.debug(f"Sucesfully executed '{result.command}'")



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2023 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import logging
from socket import gaierror
from time import sleep
from traceback import print_exc
from typing import Callable, Optional

import rpyc
from linien_common.communication import LinienControlService

from . import __version__
from .deploy import hash_username_and_password, start_remote_server
from .device import Device, generate_random_key
from .exceptions import (
    GeneralConnectionError,
    InvalidServerVersionException,
    RPYCAuthenticationException,
    ServerNotRunningException,
)
from .remote_parameters import RemoteParameters

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class ServiceWithAuth(rpyc.Service):
    def __init__(self, uuid: str, device: Device) -> None:
        super().__init__()
        self.exposed_uuid = uuid
        self.auth_hash = hash_username_and_password(
            device.username, device.password
        ).encode("utf-8")

    def _connect(self, channel, config):
        channel.stream.sock.send(self.auth_hash)  # send hash before rpyc takes over
        logger.debug("Sent authentication hash")
        return super()._connect(channel, config)


class LinienClient:
    def __init__(self, device: Device) -> None:
        """Connect to a RedPitaya that runs linien server."""
        self.device = device
        self.uuid = generate_random_key()

        # for exposing client's uuid to server
        self.client_service = ServiceWithAuth(self.uuid, self.device)

    def connect(
        self,
        autostart_server: bool,
        use_parameter_cache: bool,
        call_on_error: Optional[Callable] = None,
    ) -> None:
        self.connection = None

        i = -1
        while True:
            i += 1
            try:
                logger.info(f"Try to connect to {self.device.host}:{self.device.port}")

                self.connection = rpyc.connect(
                    self.device.host,
                    self.device.port,
                    service=self.client_service,
                    config={"allow_pickle": True},
                )

                cls = RemoteParameters
                if call_on_error:
                    cls = self._catch_network_errors(cls, call_on_error)

                self.parameters = cls(
                    self.connection.root, self.uuid, use_parameter_cache
                )

                self.control: LinienControlService = self.connection.root
                break
            except gaierror:
                # host not found
                logger.error(f"Error: host {self.device.host} not found")
                break
            except EOFError:
                logger.error("EOFError! Probably authentication failed")
                raise RPYCAuthenticationException()
            except ConnectionRefusedError:
                if not autostart_server:
                    raise ServerNotRunningException()

                if i == 0:
                    logger.error("Server is not running. Launching it!")
                    start_remote_server(self.device)
                    sleep(3)
                else:
                    if i < 20:
                        logger.info(
                            "Server still not running, waiting (may take some time)."
                        )
                        sleep(1)
                    else:
                        print_exc()
                        logger.error(
                            "Error: connection to the server could not be established"
                        )
                        break

        if self.connection is None:
            raise GeneralConnectionError()

        # now check that the remote version is the same as ours
        remote_version = self.connection.root.exposed_get_server_version().split("+")[0]
        local_version = __version__.split("+")[0]

        if (remote_version != local_version) and not ("dev" in local_version):
            raise InvalidServerVersionException(local_version, remote_version)

        self.connected = True
        logger.info("Connection established!")

    def disconnect(self) -> None:
        if self.connection is not None:
            self.connection.close()
        self.connected = False

    def _catch_network_errors(self, cls, call_on_error):
        """
        This method can be used for patching RemoteParameters such that network errors
        are redirected to `call_on_error`
        """
        function_type = type(lambda x: x)

        for attr_name in dir(cls):
            # patch all methods that don't start with __
            if not attr_name.startswith("__"):
                attr = getattr(cls, attr_name)

                if isinstance(attr, function_type):
                    method = attr

                    def wrapped(*args, method=method, **kwargs):
                        try:
                            return method(*args, **kwargs)
                        except (EOFError,):
                            logger.error("Connection lost")
                            self.connected = False
                            call_on_error()
                            raise

                    setattr(cls, attr_name, wrapped)

        return cls



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.


class RPYCAuthenticationException(Exception):
    def __init__(self):
        super().__init__(
            "Invalid credentials passed to LinienClient. Be sure to use the same "
            "username and password as when connecting via SSH."
        )


class ServerNotRunningException(Exception):
    def __init__(self):
        super().__init__(
            "The host was reached but no linien server is running. Use "
            "`autostart_server` if you want to change this."
        )


class InvalidServerVersionException(Exception):
    def __init__(self, client_version, remote_version):
        self.client_version = client_version
        self.remote_version = remote_version

        super().__init__(
            f"Version mismatch: Client is {client_version}, server is {remote_version}"
        )


class ServerNotInstalledException(Exception):
    pass


class GeneralConnectionError(Exception):
    pass
